import typing as tp
import weakref

from satella.coding import rethrow_as
from satella.coding.typing import Number

from smok.exceptions import NotReadedError
from smok.pathpoint import AdviseLevel, PathpointValueType
from smok.pathpoint.orders import Section, ReadOrder
from smok.sensor import reparse_funs
from smok.sensor.reparse import parse
from smok.sensor.types import SensorValueType, get_type


def fqtsify(tag_name: tp.Union[str, tp.Set[str]]) -> str:
    """
    Standarize the name, for use in dictionaries and other places that access sensors by name

    :param tag_name: either a space-separated set of names or a set of names proper
    :return: FQTS-ified name
    """
    if isinstance(tag_name, set):
        tag_name = list(tag_name)
    elif isinstance(tag_name, str):
        tag_name = tag_name.split(' ')
    tag_name.sort()
    return ' '.join(tag_name)


class Sensor:
    """
    A class representing a smok-client sensor defined on given device.

    :ivar fqts: a sorted string representing tags of this sensor joined by a space
    :ivar path: composite pathpoint names, separated by a tilde (~)
    :ivar type_name: name of the sensor type
    :ivar type: object used for data conversion between pathpoints and sensor values
    """
    __slots__ = ('fqts', 'path', 'type_name', '_pathpoint_names', 'type', 'device')

    def __init__(self, device: 'SMOKDevice', fqts: str, path: str, type_name: str):
        self.device = weakref.proxy(device)
        self.fqts = fqtsify(fqts)
        self.path = path
        self._pathpoint_names = path.split('~')
        self.type_name = type_name
        self.type = get_type(type_name)

    @rethrow_as(KeyError, NotReadedError)
    def _calculate_pathpoint(self, path: str) -> tp.Tuple[Number, PathpointValueType]:
        if path[0] == 'r':
            exp, pathpoints = parse(path)
            exp = exp[2:]  # skip reparse declaration and reparse point type
            pathpoint_vals = [self._calculate_pathpoint(pp) for pp in pathpoints]
            dct, ts = {}, None
            for i, pp_val in enumerate(pathpoint_vals):
                if ts is None:
                    ts = pp_val[0]
                elif pp_val[0] < ts:
                    ts = pp_val[0]
                dct['v%s' % (i,)] = pp_val[1]
            return ts, eval(exp, reparse_funs.__dict__, dct)
        else:
            pp = self.device.get_pathpoint(path)  # throws KeyError
            return pp.get()

    def get(self) -> tp.Tuple[Number, SensorValueType]:
        """
        Return the value of this sensor

        :raises NotYetReaded: required pathpoints are not available
        :return: a tuple of (timestamp, sensor value)
        :raises OperationFailedError: one of pathpoint failed to provide a value
        """
        vals = [self._calculate_pathpoint(pp) for pp in self._pathpoint_names]
        cur_ts = max(ts[0] for ts in vals)
        return cur_ts, self.type.pathpoint_to_sensor(*(val[1] for val in vals))

    def read(self, advise: AdviseLevel = AdviseLevel.ADVISE) -> Section:
        return Section([ReadOrder(y, advise) for y in self._pathpoint_names])

    @classmethod
    def from_json(cls, device: 'SMOKDevice', x: dict) -> 'Sensor':
        return Sensor(device, x['fqts'], x['path'], x['type'])



import logging
import queue
import time
import typing as tp
from concurrent.futures import wait, Future

from satella.coding import queue_get
from satella.coding.concurrent import TerminableThread, call_in_separate_thread
from satella.coding.decorators import retry

from smok.exceptions import ResponseError, NotReadedError
from smok.extras.pp_database.base import BasePathpointDatabase
from smok.pathpoint.orders import Section, WriteOrder, ReadOrder, MessageOrder, Disposition
from smok.pathpoint.pathpoint import Pathpoint

logger = logging.getLogger(__name__)


def on_read_completed_factory(oet: 'OrderExecutorThread',
                              pp: Pathpoint) -> tp.Callable[[Future], None]:
    def on_read_completed(fut: Future):
        ts = time.time()
        if fut.exception() is None:
            res = fut.result()
            if res is None:
                return
            oet.data_to_sync.on_new_data(pp.name, ts, res)
            pp.current_timestamp = time.time()
            pp.current_value = res
        else:
            exc = fut.exception()
            if isinstance(exc, NotReadedError):
                logger.error('A read future for %s returned NotReadedError, this is invalid')
                return
            exc.timestamp = ts
            oet.data_to_sync.on_new_data(pp.name, ts, exc)
            pp.current_timestamp = ts
            pp.current_value = fut.exception()

    return on_read_completed


class OrderExecutorThread(TerminableThread):
    def __init__(self, device, order_queue: queue.Queue, data_to_sync: BasePathpointDatabase):
        super().__init__(name='order executor')
        self.queue = order_queue
        self.futures_to_complete = []
        self.device = device
        self.data_to_sync = data_to_sync

    def execute_a_section(self, section: Section) -> None:
        from smok.client import SMOKDevice

        # Do we need to sync all sections?
        if section.disposition == Disposition.CANNOT_JOIN:
            self.device.sync_sections()

        if self.device.execute_section is SMOKDevice.execute_section:
            if not section.future.set_running_or_notify_cancel():
                return  # Section cancelled

            for order in section.orders:
                if isinstance(order, (WriteOrder, ReadOrder)):
                    try:
                        pathpoint = self.device.get_pathpoint(order.pathpoint)
                    except KeyError:
                        logger.warning('Got order for unavailable pathpoint %s' % (order.pathpoint,))
                        continue

                    if isinstance(order, WriteOrder):
                        if not order.is_valid():
                            continue
                        fut = pathpoint.on_write(order.value, order.advise)
                    elif isinstance(order, ReadOrder):
                        fut = pathpoint.on_read(order.advise)  # type: Future
                        fut.add_done_callback(on_read_completed_factory(self, pathpoint))

                elif isinstance(order, MessageOrder):

                    @call_in_separate_thread()
                    @retry(6, ResponseError)
                    def execute_a_message(uuid: str) -> Future:
                        self.device.api.post('/v1/device/orders/message/' + uuid)

                    fut = execute_a_message(order.uuid)
                else:
                    continue
                self.futures_to_complete.append(fut)

            time_to_wait = section.max_wait()
            if time_to_wait is not None:
                self.safe_sleep(time_to_wait)

            while self.futures_to_complete and not self.terminating:
                self.futures_to_complete = list(wait(self.futures_to_complete, 5)[1])
            section.future.set_result(None)
        else:
            self.device.execute_section(section)

    @queue_get('queue', 5)
    def loop(self, section: Section):
        print(section)
        if self.queue.qsize():
            next_section = self.queue.peek()  # type: Section
            if section.is_joinable() and next_section.is_joinable():
                section += self.queue.get()

        self.execute_a_section(section)



import typing as tp
import weakref
from abc import ABCMeta, abstractmethod
from concurrent.futures import Future

from satella.coding import wraps
from satella.coding.structures import ReprableMixin, OmniHashableMixin
from satella.coding.typing import Number

from .orders import AdviseLevel, Section, ReadOrder, WriteOrder
from .typing import PathpointValueType, ValueOrExcept
from ..basics import StorageLevel
from ..exceptions import OperationFailedError, InstanceNotReady


def must_have_device(fun):
    @wraps(fun)
    def inner(self, *args, **kwargs):
        if not self.device:
            raise InstanceNotReady('You need to attach this pathpoint to a device first')
        return fun(self, *args, **kwargs)
    return inner


class Pathpoint(ReprableMixin, OmniHashableMixin, metaclass=ABCMeta):
    """
    Base class for an user-defined pathpoint.

    Note that pathpoint is registered in the device as part of it's creation.

    :param device: device that this pathpoint should be attached to. None is also a valid option.
    :param name: pathpoint name.
    :param storage_level: storage level for this pathpoint

    :ivar name: pathpoint name
    :ivar storage_level: pathpoint's storage level
    :ivar current_timestamp: a timestamp in seconds of the last read
    :ivar current_value: last readed value or an exception instance
    :ivar device: a weak reference to the device
    """
    _HASH_FIELDS_TO_USE = ('name',)
    _REPR_FIELDS = ('name', 'storage_level')
    __slots__ = ('name', 'storage_level', 'current_value', 'current_timestamp', 'device')

    def __init__(self, device: tp.Optional['SMOKDevice'], name: str,
                 storage_level: StorageLevel = StorageLevel.TREND):
        self.device = weakref.proxy(device)
        self.name = name
        self.storage_level = storage_level
        self.current_value = None  # type: ValueOrExcept
        self.current_timestamp = None  # type: Number
        # noinspection PyProtectedMember
        if device is not None:
            device.register_pathpoint(self)

    def get_archive(self,
                    starting_at: int,
                    stopping_at: tp.Optional[int] = None) -> tp.Iterator[
            tp.Tuple[int, ValueOrExcept]]:
        """
        Get archive readings.

        This will be readed entirely from the device's data, the server will not be queried

        :param starting_at: timestamp of start, in milliseconds
        :param stopping_at: timestamp of end, in milliseconds, or None for the end of the park
        :return: an iterator of tuple (timestamp in milliseconds,
                                       pathpoint value or exception instance)
        """
        yield from self.device.pp_database.get_archive_data(self.name, starting_at, stopping_at)

    @must_have_device
    def set_new_value(self, timestamp: Number, value: ValueOrExcept) -> None:
        """
        May be called asynchronously by user threads to asynchronously update a pathpoint

        :param timestamp: new timestamp
        :param value: new value
        """
        self.device.pp_database.on_new_data(timestamp, value)
        self.current_timestamp = timestamp
        self.current_value = value

    @abstractmethod
    def on_read(self, advise: AdviseLevel) -> Future:
        """
        Called when there's a request to read this pathpoint.

        This is called from a separate thread spawned by SMOKDevice.

        The future should raise OperationFailedError when the read fails.

        .. note:: :attr:`current_timestamp` and :attr:`current_value` will be automatically
                  updated, so there's no need for the future to do that.

        :param advise: advise level of this read operation
        :returns: a Future that returns the value of this pathpoint or raises OperationFailedError.
            This Future can also return None, if the value will be set later via
            :meth:`~smokclient.pathpoint.Pathpoint.set_new_value`
        """

    @abstractmethod
    def on_write(self, value: PathpointValueType, advise: AdviseLevel) -> Future:
        """
        Called when there's a request to write this pathpoint with a particular value

        This is called from a separate thread spawned by SMOKDevice

        The future should raise OperationFailedError when the write fails.

        :param value: value to be written
        :param advise: advise level of this read operation
        :returns: a Future that completes successfully if written correctly or excepts if failed
            (you should raise OperationFailedError).
        """

    def on_new_storage_level(self, new_storage_level: StorageLevel) -> None:
        """
        Called when this pathpoint's storage level is updated.

        This is usually invoked by syncing the pathpoints with the server.

        This should complete ASAP, as this is called by a thread that is responsible
        for communication with the server.

        :param new_storage_level: new storage level for this pathpoint
        """
        self.storage_level = new_storage_level

    @must_have_device
    def get(self) -> tp.Tuple[Number, PathpointValueType]:
        """
        Return the current pathpoint value

        :return: a tuple of (last timestamp, when the operation has failed)
        :raises OperationFailedError: when pathpoint's read has failed
        """
        if self.current_value is None:
            ts, v = self.device.pp_database.get_current_value()  # raises NotReadedError
            self.current_timestamp = ts
            self.current_value = v

        if isinstance(self.current_value, OperationFailedError):
            raise self.current_value
        else:
            return self.current_timestamp, self.current_value

    def read(self, advise_level: AdviseLevel = AdviseLevel.ADVISE) -> Section:
        """
        Return an order that reads this pathpoint

        Dispatch them like this:

        >>> my_sd.execute(pp.read(2))

        :param advise_level: required advise level
        :return: a Section, whose execution will refresh the value in the pathpoint
        """
        return Section([ReadOrder(self.name, advise_level)])

    def write(self, value: PathpointValueType, advise_level: AdviseLevel = AdviseLevel.ADVISE,
              stale_after: tp.Optional[float] = None) -> Section:
        """
        Return an order, whose execution will put target value in target pathpoint.

        Dispatch them like this:

        >>> my_sd.execute(pp.write(2))

        :param value: value to write
        :param advise_level: advise level with which to write
        :param stale_after: timestamp, that if given, will prevent the write from being executed
            past it
        :return: a Section, whose execution will put the target value in the pathpoint
        """
        return Section([WriteOrder(self.name, value, advise_level, stale_after)])



from setuptools import find_packages
from distutils.core import setup
from smok import __version__


setup(version=__version__,
      url='https://github.com/smok-serwis/smok-client/',
      packages=find_packages(include=['smok', 'smok.*']),
      package_data={'smok': ['certs/dev.crt', 'certs/root.crt']},
      install_requires=['requests', 'satella>=2.14.12', 'pytz',
                        'pyasn1', 'cryptography', 'pyopenssl',
                        'ujson'],
      python_requires='!=2.7.*,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*',
      zip_safe=False
      )
